@@ -14,7 +14,7 @@ class MiningTaxesMenuItem(MenuItemHook):
         MenuItemHook.__init__(
             self,
             _("Mining Taxes"),
-            "fas fa-cube fa-fw",
+            "fas fa-cube",
             "miningtaxes:index",
             navactive=["miningtaxes:"],
         )
