@@ -1,17 +1,26 @@
 import requests
-from celery import shared_task
+from celery import chord, shared_task
 
+from django.contrib.auth.models import Permission
 from django.db import Error
 from django.http.response import Http404
 from django.shortcuts import get_object_or_404
 from django.utils import timezone
-from eveuniverse.models import EveType, EveTypeMaterial
+from esi.errors import TokenError
+from eveuniverse.models import EveGroup, EveType, EveTypeMaterial
 
 from allianceauth.eveonline.models import EveCharacter
 from allianceauth.notifications import notify
 from allianceauth.services.hooks import get_extension_logger
+from app_utils.django import users_with_permission
 
 from .app_settings import (
+    MININGTAXES_PING_CURRENT_MSG,
+    MININGTAXES_PING_CURRENT_THRESHOLD,
+    MININGTAXES_PING_FIRST_MSG,
+    MININGTAXES_PING_INTEREST_APPLIED,
+    MININGTAXES_PING_SECOND_MSG,
+    MININGTAXES_PING_THRESHOLD,
     MININGTAXES_PRICE_JANICE_API_KEY,
     MININGTAXES_PRICE_METHOD,
     MININGTAXES_PRICE_SOURCE_ID,
@@ -25,6 +34,7 @@ from .models import (
     AdminMiningCorpLedgerEntry,
     AdminMiningObsLog,
     Character,
+    CharacterMiningLedgerEntry,
     OrePrices,
     Settings,
     Stats,
@@ -63,11 +73,41 @@ def notify_taxes_due(self):
     user2taxes = calctaxes()
 
     for u in user2taxes.keys():
-        if user2taxes[u][0] > 0.01:
+        if user2taxes[u][0] > MININGTAXES_PING_THRESHOLD:
             title = "Taxes are due!"
-            message = "Please pay {:,.2f} ISK or you will be charged interest!".format(
-                user2taxes[u][0]
-            )
+            message = MININGTAXES_PING_FIRST_MSG.format(user2taxes[u][0])
+            notify(user=u, title=title, message=message, level="INFO")
+
+
+@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
+def notify_second_taxes_due(self):
+    user2taxes = calctaxes()
+
+    for u in user2taxes.keys():
+        if user2taxes[u][0] > MININGTAXES_PING_THRESHOLD:
+            title = "Taxes are due!"
+            message = MININGTAXES_PING_SECOND_MSG.format(user2taxes[u][0])
+            notify(user=u, title=title, message=message, level="INFO")
+
+
+@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
+def notify_current_taxes_threshold(self):
+    from django.contrib.auth.models import User
+
+    from .models import Stats
+
+    s = Stats.load()
+    arr = s.get_admin_main_json()
+
+    for row in arr:
+        if row["balance"] >= MININGTAXES_PING_CURRENT_THRESHOLD:
+            try:
+                u = User.objects.get(id=row["user"])
+            except Exception:
+                print(f"could not find user: {row['user']}")
+                continue
+            title = "Taxes are due!"
+            message = MININGTAXES_PING_CURRENT_MSG.format(row["balance"])
             notify(user=u, title=title, message=message, level="INFO")
 
 
@@ -80,33 +120,50 @@ def apply_interest(self):
         if user2taxes[u][0] <= 0.01:
             continue
         interest = round(user2taxes[u][0] * settings.interest_rate / 100.0, 2)
-        if interest > 0.01:
+        if interest > MININGTAXES_PING_THRESHOLD:
             user2taxes[u][2].give_credit(-1.0 * interest, "interest")
             title = "Taxes are overdue!"
-            message = (
-                "An interest of {:,.2f} ISK has been charged for late taxes.".format(
-                    interest
-                )
-            )
+            message = MININGTAXES_PING_INTEREST_APPLIED.format(interest)
             notify(user=u, title=title, message=message, level="WARN")
 
 
 @shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
 def update_daily(self):
+    logger.info("Beginning price update")
     update_all_prices()
+
+    logger.info("Beginning admin character updates")
     characters = AdminCharacter.objects.all()
     for character in characters:
         update_admin_character(character_pk=character.id, celery=True)
+
+    logger.info("Running character update in parallel")
+    chartasks = []
     characters = Character.objects.all()
     for character in characters:
-        update_character(character_pk=character.id, celery=True)
+        chartasks.append(update_character.s(character_pk=character.id, celery=True))
+
+    chord(chartasks)(precalcs.s())
+
+
+@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
+def precalcs(self, output):
+    perc = round(sum(output) / len(output) * 100, 2)
+    logger.info(f"Starting Precalc: {perc}% update success")
+    logger.info("Adding Corp Moon Taxes")
     add_corp_moon_taxes()
+    logger.info("Linking Tax Payments")
     add_tax_credits()
+    logger.info("Running each character's precalc")
+    characters = Character.objects.all()
     # precalc all characters
     for character in characters:
         character.precalc_all()
+
+    logger.info("Running Stats Precalc")
     s = Stats.load()
     s.precalc_all()
+    logger.info("Finished precalc, updates complete")
 
 
 def valid_janice_api_key():
@@ -162,12 +219,26 @@ def get_bulk_prices(type_ids):
 
 @shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
 def update_all_prices(self):
+
     type_ids = []
     market_data = {}
     api_up = True
 
-    # Get all type ids
-    prices = PriceGroups().items
+    # Get all type ids by known groups
+    prices = []
+    added = set()
+    for grp in PriceGroups().groups:
+        EveGroup.objects.get_or_create_esi(
+            id=grp,
+            include_children=True,
+            wait_for_children=True,
+        )
+        g = EveType.objects.filter(eve_group_id=grp)
+        for it in g:
+            if it.id in added:
+                continue
+            added.add(it.id)
+            prices.append(it)
 
     # Update EveUniverse objects
     matset = set()
@@ -387,16 +458,27 @@ def add_corp_moon_taxes_by_char(character):
         miner_id=character.eve_character.character_id
     )
     for entry in entries:
-        (row, _) = character.mining_ledger.update_or_create(
-            date=entry.date,
-            eve_solar_system=entry.eve_solar_system,
-            eve_type=entry.eve_type,
-            defaults={"quantity": entry.quantity},
-        )
-        row.calc_prices()
+        try:
+            row = character.mining_ledger.get(
+                date=entry.date,
+                eve_solar_system=entry.eve_solar_system,
+                eve_type=entry.eve_type,
+            )
+            if row.quantity != entry.quantity:
+                row.quantity = entry.quantity
+                row.save()
+                row.calc_prices()
+        except CharacterMiningLedgerEntry.DoesNotExist:
+            row = character.mining_ledger.create(
+                date=entry.date,
+                eve_solar_system=entry.eve_solar_system,
+                eve_type=entry.eve_type,
+                quantity=entry.quantity,
+            )
+            row.calc_prices()
 
 
-@shared_task(**{**TASK_DEFAULT_KWARGS, **{"bind": True}})
+@shared_task(ignore_result=False, **{**TASK_DEFAULT_KWARGS, **{"bind": True}})
 def update_character(
     self, character_pk: int, force_update: bool = False, celery=False
 ) -> bool:
@@ -421,11 +503,24 @@ def update_character(
         logger.info("%s: No update required", character)
         return False
 
-    logger.info(
-        "%s: Starting %s character update", character, "forced" if force_update else ""
-    )
-
-    character.update_mining_ledger()
-    add_tax_credits_by_char(character)
-    if not celery and MININGTAXES_TAX_ONLY_CORP_MOONS:
-        add_corp_moon_taxes_by_char(character)
+    try:
+        logger.info(
+            "%s: Starting %s character update",
+            character,
+            "forced" if force_update else "",
+        )
+        character.update_mining_ledger()
+        add_tax_credits_by_char(character)
+        if not celery and MININGTAXES_TAX_ONLY_CORP_MOONS:
+            add_corp_moon_taxes_by_char(character)
+    except TokenError:
+        p = Permission.objects.get(
+            content_type__app_label=Character._meta.app_label, codename="admin_access"
+        )
+        logger.warn("Missing ESI Token for %s", character)
+        title = f"Missing ESI Token for {character}"
+        message = f"MiningTaxes could not fetch the mining ledger for {character}."
+        for u in users_with_permission(p):
+            notify(user=u, title=title, message=message, level="CRITICAL")
+        pass
+    return False
